import pathlib
import tempfile

from mpl_data_cast import Recipe


def make_example_data():
    td = pathlib.Path(tempfile.mkdtemp())
    d1 = td / "hans" / "peter"
    d1.mkdir(parents=True, exist_ok=True)
    d2 = td / "fliege"
    d2.mkdir(parents=True, exist_ok=True)
    (d1 / "a.txt").write_text("hello ")
    (d1 / "b.txt").write_text("world!")
    (d2 / "1.txt").write_text("lorem ")
    (d2 / "2.txt").write_text("ipsum ")
    (d2 / "3.txt").write_text("dolor sit ")
    (d2 / "4.txt").write_text("amet.")
    return td


class DummyRecipe(Recipe):
    """A pipeline that just concatenates text files"""
    def convert_dataset(self, path_list, temp_path, **kwargs):
        data = ""
        for pp in path_list:
            data += pp.read_text()
        temp_path.write_text(data)

    def get_raw_data_iterator(self):
        """Return a sorted list of files"""
        data_list = []
        for pp in self.path_raw.rglob("*"):
            if pp.is_dir():
                files = sorted(pp.rglob("*.txt"))
                if files:
                    data_list.append(files)
        return sorted(data_list)


def test_pipeline_init():
    path_raw = make_example_data()
    path_tar = pathlib.Path(tempfile.mkdtemp()) / "test"
    pl = DummyRecipe(path_raw, path_tar)
    assert pl.tempdir.exists()
    assert pl.format == "DummyRecipe"
    assert not path_tar.exists()


def test_pipeline_cast():
    path_raw = make_example_data()
    path_tar = pathlib.Path(tempfile.mkdtemp()) / "test"
    pl = DummyRecipe(path_raw, path_tar)
    pl.cast()
    text1 = (path_tar / "hans" / "peter" / "a.txt").read_text()
    assert text1 == "hello world!"
    text2 = (path_tar / "fliege" / "1.txt").read_text()
    assert text2 == "lorem ipsum dolor sit amet."


def test_pipeline_cast_delete_after():
    path_raw = make_example_data()
    path_tar = pathlib.Path(tempfile.mkdtemp()) / "test"
    pl = DummyRecipe(path_raw, path_tar)
    input_list = sorted(pl.get_raw_data_iterator())
    output_list = [pl.get_target_path(pi) for pi in input_list]
    for po in output_list:
        assert not po.exists()
    ret = pl.cast()
    assert ret["success"]
    for po in output_list:
        assert po.exists()
    # This is the actual important test. The instance should remove
    # temporary files as it transfers them.
    temp_files = sorted(pl.tempdir.rglob("*.txt"))
    assert len(temp_files) == 0


def test_pipeline_get_target_path():
    path_raw = make_example_data()
    path_tar = pathlib.Path(tempfile.mkdtemp()) / "test"
    path_tar.mkdir()
    pl = DummyRecipe(path_raw, path_tar)

    tar1 = path_tar / "fliege" / "1.txt"
    assert str(pl.get_target_path(pl.get_raw_data_iterator()[0])) == str(tar1)

    tar2 = path_tar / "hans" / "peter" / "a.txt"
    assert str(pl.get_target_path(pl.get_raw_data_iterator()[1])) == str(tar2)



import h5py

import re


def is_valid_h5_file(path):
    try:
        with h5py.File(path):
            valid = True
    except BaseException:
        valid = False
    return valid


def h5_file_contains(path, location="/", attribute=None, verifier=None,
                     regexp=None):
    """Check whether the HDF5 file contains certain attributes or datastes

    Parameters
    ----------
    path: str or pathlib.Path
        path to the HDF5 file
    location: str
        location in the HDF5 structure (can be a dataset or a group)
    attribute: str
        the attribute to check
    verifier: callable
        function to check against
    regexp: str
        regular expression pattern to check against (redundant to verifier)
        using `re.fullmatch(regexp, str(h5[location].attrs[attribite]))`
    """
    contains = False
    with h5py.File(path) as h5:
        if location in h5:
            loc = h5[location]
            if attribute is None:
                # we are only checking for a location
                contains = True
                if verifier is not None:
                    raise ValueError(
                        "`verifier` specified without `attribute`!")
                if regexp is not None:
                    raise ValueError(
                        "`regexp` specified without `attribute`!")
            if attribute is not None and attribute in loc.attrs:
                attr = loc.attrs[attribute]
                if verifier is not None:
                    verified = verifier(attr)
                else:
                    verified = True
                if regexp is not None:
                    matched = re.fullmatch(regexp, str(attr)) is not None
                else:
                    matched = True
                contains = verified and matched
    return contains



"""Screenshots for documentation """
import sys
from PyQt6.QtWidgets import QApplication

from mpl_data_cast.gui.main import MPLDataCast
from mpl_data_cast.gui.preferences import Preferences

app = QApplication(sys.argv)

mw = MPLDataCast()
mw.grab().save("ui_main.png")

prf = Preferences(mw)
prf.grab().save("ui_preferences.png")

mw.close()



"""basic tests"""
from unittest import mock

from PyQt6 import QtCore, QtTest, QtWidgets

import mpl_data_cast
from mpl_data_cast.gui.main import MPLDataCast
from mpl_data_cast.gui.widget_output import OutputWidget
from mpl_data_cast.gui.widget_input import InputWidget


def test_simple(qtbot):
    """Open the main window and close it again, check that some basic
    attributes exist."""
    mw = MPLDataCast()
    qtbot.addWidget(mw)
    QtWidgets.QApplication.setActiveWindow(mw)
    QtTest.QTest.qWait(100)

    assert isinstance(mw.widget_input, InputWidget)
    assert isinstance(mw.widget_output, OutputWidget)
    assert isinstance(mw.pushButton_transfer, QtWidgets.QPushButton)
    mw.close()
    QtTest.QTest.qWait(100)
    QtWidgets.QApplication.processEvents(
        QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 5000)


def test_on_action_about(qtbot):
    """Check that the 'About' menu works."""
    with mock.patch("PyQt6.QtWidgets.QMessageBox.about") as mock_about:
        mw = MPLDataCast()
        qtbot.addWidget(mw)
        QtWidgets.QApplication.setActiveWindow(mw)
        QtTest.QTest.qWait(100)
        mw.on_action_about()
        mw.close()
        QtTest.QTest.qWait(100)
        QtWidgets.QApplication.processEvents(
            QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 5000)

        assert mock_about.call_args.args[1] == \
               f"MPL-Data-Cast {mpl_data_cast.__version__}"
        assert "MPL-Data-Cast" in mock_about.call_args.args[2]



import json
import warnings
import pathlib
import h5py
import numpy as np
import tifffile

from ..recipe import Recipe


meta_data_mapping = {
    "pixel size": ["pixel_size", "PixelSizeUm", lambda x: x * 1e-6],
    "time": ["UNUSED", "ElapsedTime-ms", lambda x: x * 1e3],
    "date": ["UNUSED", "ReceivedTime", lambda x: x.split()[0]],
    "identifier": ["UNUSED", "UUID", lambda x: x],
    "pos x": ["UNUSED", "XPositionUm", lambda x: x * 1e-6],
    "pos y": ["UNUSED", "YPositionUm", lambda x: x * 1e-6],
    "focus": ["UNUSED", "ZPositionUm", lambda x: x * 1e-6],
    "wavelength": ["wavelength", "QLSIIllumFilter-Label",
                   # "QLSIIllumFilter-FBH488-10-CWL488-FWHM10-St483-Stp493"
                   lambda x: float(x.split("-")[3].strip("CWL"))*1e-9],
    # TODO
    "medium index": ["medium_index", "RI_UNDEFINED", lambda x: x],
    "numerical aperture": ["UNUSED", "NA_UNDEFINED", lambda x: x],
}


class QLSIRecipe(Recipe):
    """ome.tif file format from MicroManager with Phasics SID4Bio camera"""
    def _write_h5_dataset_metadata(self, path, ds, json_meta_data=None,
                                   warn=True):
        if json_meta_data is None:
            json_meta_data = {}
        for key in meta_data_mapping:
            parname, spim_name, converter = meta_data_mapping[key]
            # check if the attributes are set
            kw_ds = ds.attrs.get(key, None)
            if kw_ds is None:
                # fill in missing metadata from json dictionary
                mval = json_meta_data.get(spim_name, None)
                if mval is None:
                    if warn:
                        warnings.warn(f"No {key} defined for {path}!")
                else:
                    ds.attrs[key] = converter(mval)

        # Create and Set image attributes:
        # HDFView recognizes this as a series of images.
        # Use np.string_ as per
        # http://docs.h5py.org/en/stable/strings.html#compatibility
        ds.attrs.create('CLASS', np.string_('IMAGE'))
        ds.attrs.create('IMAGE_VERSION', np.string_('1.2'))
        ds.attrs.create('IMAGE_SUBCLASS', np.string_('IMAGE_GRAYSCALE'))

    def convert_dataset(self, path_list: list, temp_path: pathlib.Path,
                        wavelength: float = None,
                        pixel_size: float = None,
                        medium_index: float = None,
                        qlsi_pitch_term: float = 1.87711e-08
                        ):
        """Convert QLSI TIF data to qpformat HDF5 format"""
        # get the metadata
        meta_text = path_list[1].read_text(errors="ignore")
        meta_data_full = json.loads(meta_text)

        # extract the information from every frame
        meta_data_list = []
        for key in meta_data_full:
            if key.startswith("FrameKey"):
                meta_data_list.append(meta_data_full[key])
        meta_data_list = sorted(meta_data_list,
                                key=lambda x: x["ElapsedTime-ms"])

        # get the data from the tif file
        data = []
        with tifffile.TiffFile(path_list[0]) as tif:
            for ii in range(len(tif.pages)):
                data.append(tif.asarray(ii))

        # get the reference data
        ref_data = tifffile.imread(str(path_list[2]))

        # Sanity checks
        slices = meta_data_full["Summary"].get("Slices", 1)
        frames = meta_data_full["Summary"].get("Frames", 1)
        # (positions are individual files)
        # (channels are not handled here)
        if frames * slices != len(data):
            raise ValueError("Size mismatch in data and meta data!")

        # prepare HDF5 metadata
        dsattrs = {}
        local_parms = locals()
        for key in meta_data_mapping:
            parname, spim_name, converter = meta_data_mapping[key]
            # get keyword argument given to this function
            kw_val = local_parms.get(parname, None)
            if kw_val is not None:  # e.g. `if wavelength is not None`
                dsattrs[key] = kw_val

        if qlsi_pitch_term:
            dsattrs["qlsi_pitch_term"] = qlsi_pitch_term

        dsattrs["device"] = meta_data_full["Summary"]["ComputerName"]
        dsattrs["software"] = "MicroManager " \
            + meta_data_full["Summary"]["MicroManagerVersion"]

        # Create the HDF5 file
        with h5py.File(temp_path, "w") as h5:
            # Store the entire meta data file
            write_text_dataset(h5.require_group("logs"),
                               "meta_data",
                               meta_text.split("\n"))

            # Store the reference meta data file
            write_text_dataset(
                h5.require_group("logs"),
                "meta_data_ref",
                path_list[3].read_text(errors="ignore").split("\n"))

            # write qpformat metadata identifier
            h5.attrs["file_format"] = "qpformat"
            h5.attrs["imaging_modality"] = \
                "quadriwave lateral shearing interferometry"

            # Write reference data
            ds = h5.create_dataset(
                name="reference",
                chunks=ref_data.shape,
                data=ref_data,
                fletcher32=True,
                compression=True,
                compression_opts=9,
            )
            ds.attrs.update(dsattrs)
            self._write_h5_dataset_metadata(
                path=path_list[2],
                ds=ds,
                warn=False,
            )

            # Write series data
            for ii, img in enumerate(data):
                json_meta_data = meta_data_list[ii]
                ds = h5.create_dataset(
                    name=str(ii),
                    chunks=img.shape,
                    data=img,
                    fletcher32=True,
                    compression=True,
                    compression_opts=9,
                )
                ds.attrs.update(dsattrs)
                self._write_h5_dataset_metadata(
                    path=path_list[0],
                    ds=ds,
                    json_meta_data=json_meta_data
                )

    def get_raw_data_iterator(self):
        """ Get raw data files

        Data directory structure for measurement and reference (ref) data.

        - QLSI_2022-04-29_ZStack_Example_1
          - comments.txt
          - DisplaySettings.json
          - QLSI_2022-04-26_ZStack_Example_1_MMStack_Pos0_metadata.txt
          - QLSI_2022-04-26_ZStack_Example_1_MMStack_Pos0.ome.tif
        - QLSI_2022-04-29_ZStack_Example_ref_1
          - comments.txt
          - DisplaySettings.json
          - QLSI_2022-04-26_ZStack_Example_ref_1_MMStack_Pos0_metadata.txt
          - QLSI_2022-04-26_ZStack_Example_ref_1_MMStack_Pos0.ome.tif
        """
        # This is from the preliminary data
        for pp in sorted(self.path_raw.rglob("*.tif")):
            if pp.name.count("_ref_"):
                # ignore reference measurements
                continue
            if self.is_valid_file(pp):
                meta_path = pp.with_name(pp.name[:-8] + "_metadata.txt")
                name_stem, num = pp.parent.name.rsplit("_", 1)
                ref_dir = pp.parent.with_name(f"{name_stem}_ref_{num}")
                ref_pp = list(ref_dir.glob("*.tif"))[0]
                ref_meta = ref_pp.with_name(ref_pp.name[:-8] + "_metadata.txt")
                # handle junk data (included, so we don't copy it)
                junk = []
                for pi in [pp, ref_pp]:
                    junk.append(pi.parent / "comments.txt")
                    junk.append(pi.parent / "DisplaySettings.json")
                yield [pp, meta_path, ref_pp, ref_meta] + junk

    def get_target_path(self, path_list):
        """Get the target path .h5 for a path_list"""
        target_p = super(QLSIRecipe,
                         self).get_target_path(path_list)
        name = target_p.name[:-8] + ".h5"
        # get rid of one directory hierarchy level
        return target_p.parent.parent / name

    @staticmethod
    def is_valid_file(path):
        valid = False
        if path.name.endswith(".ome.tif"):
            meta_path = path.with_name(path.name[:-8] + "_metadata.txt")
            # also check for reference measurement
            name_stem, num = path.parent.name.rsplit("_", 1)
            ref_dir = path.parent.with_name(f"{name_stem}_ref_{num}")
            if (ref_dir.exists()
                and bool(list(ref_dir.glob("*.ome.tif")))
                    and bool(list(ref_dir.glob("*_metadata.txt")))
                    and meta_path.exists()
                    and meta_path.read_text(
                        errors="ignore").count("QLSICamera")):
                valid = True
        return valid


def write_text_dataset(group, name, lines):
    """Write text to an HDF5 dataset

    Text data are written as a fixed-length string dataset.

    Parameters
    ----------
    group: h5py.Group
        parent group
    name: str
        name of the dataset containing the text
    lines: list of str or str
        the text, line by line
    """
    # replace text
    if name in group:
        del group[name]

    lnum = len(lines)
    # Determine the maximum line length and use fixed-length strings,
    # because compression and fletcher32 filters won't work with
    # variable length strings.
    # https://github.com/h5py/h5py/issues/1948
    # 100 is the recommended maximum and the default, because if
    # `mode` is e.g. "append", then this line may not be the longest.
    max_length = 100
    lines_as_bytes = []
    for line in lines:
        # convert lines to bytes
        if not isinstance(line, bytes):
            lbytes = line.encode("UTF-8")
        else:
            lbytes = line
        max_length = max(max_length, len(lbytes))
        lines_as_bytes.append(lbytes)

    # Create the dataset
    txt_dset = group.create_dataset(
        name,
        shape=(lnum,),
        dtype=f"S{max_length}",
        maxshape=(None,),
        chunks=True,
        fletcher32=True,
        compression="gzip",
        compression_opts=9)

    # Write the text data line-by-line
    for ii, lbytes in enumerate(lines_as_bytes):
        txt_dset[ii] = lbytes



# flake8: noqa: F401
# import recipes here so they get are defined and available
from .rcp_catchall import CatchAllRecipe
from .rcp_oah import OAHRecipe
from .rcp_qlsi import QLSIRecipe
from .rcp_rtdc import RTDCRecipe



import tempfile as tf
import pathlib
import shutil

import pytest

from helper import retrieve_data
from mpl_data_cast.path_tree import PathTree, PathError


def test_simple_tree_from_dir():
    """Check basic attributes of a small tree"""
    with tf.TemporaryDirectory() as tmpdir:
        tmpdir = pathlib.Path(tmpdir)
        tree = PathTree(tmpdir, depth_limit=10)
        assert isinstance(tree.tree_root, pathlib.Path)
        assert isinstance(tree.children, dict)
        assert tree.tree_depth == 1


def test_simple_tree_from_file():
    """Check PathTree generation with just one file."""
    with tf.TemporaryDirectory() as tmpdir:
        with tf.NamedTemporaryFile(dir=tmpdir) as tmpfile:
            tmpfile = pathlib.Path(str(tmpfile.name))
            tree = PathTree(tmpfile)
            assert isinstance(tree.tree_root, pathlib.Path)
            assert tree.tree_depth == 1
            assert tree.tree_root == tmpfile.parent


def test_wrong_input_raises_errors(tmp_path):
    """Check that errors are raised in case the input for the PathTree
    constructor is wrong."""
    with pytest.raises(TypeError):
        _ = PathTree(r"some/path/to/dir")
    with pytest.raises(PathError):
        _ = PathTree(pathlib.Path(r"some/path/somewhere"))


def test_tree_with_subdir():
    """Check PathTree generation with a subfolder"""
    with tf.TemporaryDirectory() as tmpdir:
        with tf.TemporaryDirectory(dir=tmpdir) as tmpdir2:
            tmpdir = pathlib.Path(tmpdir)
            tmpdir2 = pathlib.Path(tmpdir2)

            tree = PathTree(tmpdir)
            assert tree.tree_depth == 2
            assert len(tree.children) == 1
            assert tmpdir2.name in tree.children.keys()
            assert isinstance(tree.children[tmpdir2.name], PathTree)


def test_tree_with_subdir_and_files(tmp_path):
    path_in = retrieve_data("rcp_rtdc_mask-contour_2018.zip")
    test_file = path_in / "M001_data.rtdc"
    t_file1 = tmp_path / "M001_data.rtdc"
    shutil.copy(test_file, t_file1)
    with tf.TemporaryDirectory(dir=tmp_path) as tmpdir2:
        tmpdir2 = pathlib.Path(tmpdir2)
        t_file2 = tmpdir2 / "M002_data.rtdc"
        shutil.copy(t_file1, t_file2)

        assert t_file1.exists()
        assert t_file2.exists()
        tree = PathTree(tmp_path)

        assert tree.tree_depth == 2
        assert len(tree.get_file_list()) == 1
        child_tree = tree.children[tmpdir2.name]
        assert isinstance(child_tree, PathTree)
        assert child_tree.tree_depth == 1
        assert child_tree.get_file_list()[0] == t_file2


def test_tree_with_hidden_file(tmp_path):
    """Check that PathTree ignores hidden files which start with a '.'"""
    path_in = retrieve_data("rcp_rtdc_mask-contour_2018.zip")
    test_file = path_in / "M001_data.rtdc"
    t_file1 = tmp_path / "M001_data.rtdc"
    # and a "hidden" file
    t_file2 = tmp_path / ".M001_data.rtdc"
    shutil.copy(test_file, t_file1)
    shutil.copy(test_file, t_file2)
    with tf.TemporaryDirectory(dir=tmp_path) as tmpdir2:
        tmpdir2 = pathlib.Path(tmpdir2)
        t_file3 = tmpdir2 / "M002_data.rtdc"
        shutil.copy(t_file1, t_file3)

        assert t_file1.exists()
        assert t_file2.exists()
        assert t_file3.exists()
        tree = PathTree(tmp_path)

        assert tree.tree_depth == 2
        assert len(tree.get_file_list()) == 1
        child_tree = tree.children[tmpdir2.name]
        assert isinstance(child_tree, PathTree)
        assert child_tree.tree_depth == 1
        assert child_tree.get_file_list()[0] == t_file3


def test_path_in_tree(tmp_path):
    """Check that the __contains__ functionality works"""
    path_in = retrieve_data("rcp_rtdc_mask-contour_2018.zip")
    test_file = path_in / "M001_data.rtdc"
    t_file1 = tmp_path / "M001_data.rtdc"
    shutil.copy(test_file, t_file1)
    with tf.TemporaryDirectory(dir=tmp_path) as tmpdir2:
        tmpdir2 = pathlib.Path(tmpdir2)
        with tf.TemporaryDirectory(dir=tmpdir2) as tmpdir3:
            tmpdir3 = pathlib.Path(tmpdir3)

            tree = PathTree(tmp_path)
            assert tree.tree_depth == 3
            assert tmpdir2 in tree
            assert tmpdir3 in tree


def test_retrieve_full_tree(tmp_path):
    """Check that the function `retrieve_full_path_tree()` acutally returns
    a PathTree object with all subdirectories in full depth."""
    with tf.TemporaryDirectory(dir=tmp_path) as tmpdir2:
        tmpdir2 = pathlib.Path(tmpdir2)
        with tf.TemporaryDirectory(dir=tmpdir2) as tmpdir3:
            _ = pathlib.Path(tmpdir3)

            tree = PathTree(tmp_path, depth_limit=2)
            assert tree.tree_depth == 2

            full_tree = tree.retrieve_full_path_tree()
            assert full_tree.tree_depth == 3
            assert full_tree.tree_root.samefile(tree.tree_root)


def test_small_path_tree(tmp_path):
    """Have a shallow PathTree object and check that the children on the
    deepest level do not have PathTree objects as children, but only a list
    of subdirectories in `PathTree.list_child_dirs`."""
    with tf.TemporaryDirectory(dir=tmp_path) as tmpdir2:
        tmpdir2 = pathlib.Path(tmpdir2)
        with tf.TemporaryDirectory(dir=tmpdir2) as tmpdir3:
            tmpdir3 = pathlib.Path(tmpdir3)

            tree = PathTree(tmp_path, depth_limit=2)
            assert tree.tree_depth == 2
            child_tree = tree.children[tmpdir2.name]
            assert not child_tree.children
            assert child_tree.list_child_dirs
            assert child_tree.list_child_dirs[0] == tmpdir3.name

            # check whether the __contains__ magic function was implemented
            # correctly
            assert tmpdir2 in tree
            assert tmpdir3 in tree
            assert tmpdir3 in child_tree



import numpy as np

from helper import retrieve_data

from mpl_data_cast.mod_recipes import RTDCRecipe

import dclab


def test_pipeline_cast_convert_error(tmp_path):
    path_in = retrieve_data("rcp_rtdc_mask-contour_2018.zip")
    name = "M002_data.rtdc"
    (path_in / name).touch()  # an invalid rtdc file

    rcp = RTDCRecipe(path_raw=path_in, path_tar=tmp_path)
    result = rcp.cast()
    assert not result["success"]
    assert len(result["errors"]) == 1
    assert name in str(result["errors"][0][0])
    assert not (tmp_path / name).exists()


def test_rcp_rtdc_base(tmp_path):
    path_in = retrieve_data("rcp_rtdc_mask-contour_2018.zip")

    rcp = RTDCRecipe(path_raw=path_in, path_tar=tmp_path)
    rcp.cast()

    assert (path_in / "M001_SoftwareSettings.ini").exists()
    assert not (tmp_path / "M001_SoftwareSettings.ini").exists()

    pin = path_in / "M001_data.rtdc"
    pout = tmp_path / "M001_data.rtdc"

    with dclab.new_dataset(pin) as ds1, dclab.new_dataset(pout) as ds2:
        assert "M001_SoftwareSettings.ini" not in ds1.logs
        assert "M001_SoftwareSettings.ini" in ds2.logs
        assert np.all(ds1["deform"] == ds2["deform"])

    # Make sure compression worked (even with the ini file in it)
    assert pin.stat().st_size > pout.stat().st_size


def test_rcp_rtdc_with_other_files(tmp_path):
    path_in = retrieve_data("rcp_rtdc_mask-contour_2018.zip")
    # an additional file that should be copied
    (path_in / "additional.docx").touch()

    rcp = RTDCRecipe(path_raw=path_in, path_tar=tmp_path)
    rcp.cast()

    assert (tmp_path / "additional.docx").exists()



import os
import time
import signal
import sys
import traceback
import pathlib
import dclab
import h5py
import numpy
import pkg_resources
from PyQt6 import uic, QtCore, QtGui, QtWidgets

from ..recipe import IGNORED_FILE_NAMES
from ..mod_recipes.rcp_rtdc import RTDCRecipe
from .._version import __version__

from . import preferences
from . import splash


# set Qt icon theme search path
QtGui.QIcon.setThemeSearchPaths([
    os.path.join(pkg_resources.resource_filename("mpl_data_cast", "img"),
                 "icon-theme")])
QtGui.QIcon.setThemeName(".")


class MPLDataCast(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        """Initialize MPL-Data-Cast"""
        QtCore.QCoreApplication.setOrganizationName("MPL")
        QtCore.QCoreApplication.setApplicationName("MPL-Data-Cast")
        super(MPLDataCast, self).__init__(*args, **kwargs)

        # if "--version" was specified, print the version and exit
        if "--version" in sys.argv:
            print(__version__)
            QtWidgets.QApplication.processEvents(
                QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 300)
            sys.exit(0)

        path_ui = pkg_resources.resource_filename("mpl_data_cast.gui",
                                                  "main.ui")
        uic.loadUi(path_ui, self)

        # settings
        QtCore.QSettings.setDefaultFormat(QtCore.QSettings.Format.IniFormat)
        self.settings = QtCore.QSettings()
        # self.settings.setIniCodec("utf-8")

        # load some values from the settings
        self.widget_output.tree_depth_limit = int(self.settings.value(
            "rtdc/tree_depth_limit", 3))
        self.widget_input.tree_depth_limit = int(self.settings.value(
            "rtdc/tree_depth_limit", 3))
        self.widget_output.update_output_dir(
            self.settings.value("rtdc/output_path", str(pathlib.Path.cwd())))
        # signals
        self.pushButton_transfer.clicked.connect(self.on_task_transfer)
        # GUI
        self.setWindowTitle(f"MPL-Data-Cast {version}")
        # Disable native menu bar (e.g. on Mac)
        self.menubar.setNativeMenuBar(False)
        # File menu
        self.actionPreferences.triggered.connect(self.on_action_preferences)
        self.actionQuit.triggered.connect(self.on_action_quit)
        # Help menu
        self.actionSoftware.triggered.connect(self.on_action_software)
        self.actionAbout.triggered.connect(self.on_action_about)

        self.show()
        self.raise_()
        splash.splash_close()

    @QtCore.pyqtSlot()
    def on_action_preferences(self):
        """Show the preferences dialog"""
        prev_tree_depth_limit = self.settings.value("rtdc/tree_depth_limit", 8)
        dlg = preferences.Preferences(self)
        dlg.setWindowTitle("MPL-Data-Cast Preferences")
        dlg.exec()
        # update maximum tree depth if necessary
        if self.settings.value("rtdc/tree_depth_limit",
                               8) != prev_tree_depth_limit:
            self.widget_output.tree_depth_limit = int(self.settings.value(
                "rtdc/tree_depth_limit", 8))
            self.widget_input.tree_depth_limit = int(self.settings.value(
                "rtdc/tree_depth_limit", 8))
            if self.widget_output.path is not None:
                self.widget_output.update_output_dir(self.widget_output.path)
            if self.widget_input.path is not None:
                self.widget_input.update_input_dir(self.widget_input.path)

    @QtCore.pyqtSlot()
    def on_action_quit(self) -> None:
        """Determine what happens when the user wants to quit"""
        QtCore.QCoreApplication.quit()

    def on_action_about(self) -> None:
        """Show imprint."""
        gh = "GuckLab/MPL-Data-Cast"
        rtd = "mpl-data-cast.readthedocs.io"
        about_text = "Convert and transfer data.<br><br>" \
                     + "Author: Paul Müller and others<br>" \
                     + "GitHub: " \
                     + "<a href='https://github.com/{gh}'>{gh}</a><br>".format(gh=gh) \
                     + "Documentation: " \
                     + "<a href='https://{rtd}'>{rtd}</a><br>".format(rtd=rtd)  # noqa 501
        QtWidgets.QMessageBox.about(self,
                                    "MPL-Data-Cast {}".format(version),
                                    about_text)

    @QtCore.pyqtSlot()
    def on_action_software(self) -> None:
        """Show used software packages and dependencies."""
        libs = [dclab,
                h5py,
                numpy,
                ]

        sw_text = f"MPL-Data-Cast {version}\n\n"
        sw_text += f"Python {sys.version}\n\n"
        sw_text += "Modules:\n"
        for lib in libs:
            sw_text += f"- {lib.__name__} {lib.__version__}\n"
        sw_text += f"- PyQt6 {QtCore.QT_VERSION_STR}\n"

        QtWidgets.QMessageBox.information(self, "Software", sw_text)

    @QtCore.pyqtSlot()
    def on_task_transfer(self) -> None:
        """Execute recipe to transfer data."""
        if not self.widget_input.path.exists():
            QtWidgets.QMessageBox.information(self, "Error",
                                              "Input directory not correct!")
        if not self.widget_output.path.exists():
            QtWidgets.QMessageBox.information(self, "Error",
                                              "Output directory not correct!")

        rp = RTDCRecipe(self.widget_input.path, self.widget_output.path)

        nb_files = 0  # counter for files, used for progress bar
        for elem in self.widget_input.path.rglob("*"):
            if elem.is_file() and elem.name not in IGNORED_FILE_NAMES:
                nb_files += 1
        with Callback(self, nb_files) as path_callback:
            result = rp.cast(path_callback=path_callback)
        if result["success"]:
            QtWidgets.QMessageBox.information(self, "Transfer completed",
                                              "Data transfer completed.")
            self.progressBar.setValue(0)
            QtWidgets.QApplication.processEvents(
                QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 300)
            self.widget_output.update_tree()
        else:
            msg = "Some problems occured during data transfer:\n"
            for path, _ in result["errors"]:
                msg += f" - {path}\n"

            QtWidgets.QMessageBox.information(self, "Error", msg)

            text = ""
            for path, tb in result["errors"]:
                text += f"PATH {path}:\n{tb}\n\n"
            pathlib.Path("mpldc-dump.txt").write_text(text)


class Callback:
    """Makes it possible to execute code everytime a file was processed.
    Used for updating the progress bar and calculating the processing rate."""

    def __init__(self, gui, max_count: int):
        self.gui = gui
        self.counter = 0
        self.max_count = max_count
        self.size = 0
        self.time_start = time.monotonic()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __call__(self, path) -> None:
        self.counter += 1
        self.size += path.stat().st_size
        self.gui.progressBar.setValue(int(self.counter / self.max_count * 100))
        QtWidgets.QApplication.processEvents(
            QtCore.QEventLoop.ProcessEventsFlag.AllEvents, 300)

    def get_rate(self) -> float:
        """Calculate processing rate in MB/s"""
        curtime = time.monotonic()
        if curtime > self.time_start:
            return self.size / 1024 ** 2 / (curtime - self.time_start)
        else:
            return 0


def excepthook(etype, value, trace) -> None:
    """
    Handler for all unhandled exceptions.

    Parameters
    ----------
    etype : Type[BaseException]
        the exception type (`SyntaxError`, `ZeroDivisionError`, etc...)
    value : BaseException | str
        The exception error message;
    trace : TracebackType | str
        the traceback header, if any (otherwise, it prints the standard
        Python header: ``Traceback (most recent call last)``.
    """
    vinfo = f"Unhandled exception in MPL-Data-Cast version {version}:\n"
    tmp = traceback.format_exception(etype, value, trace)
    exception = "".join([vinfo] + tmp)

    errorbox = QtWidgets.QMessageBox()
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.ButtonRole.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.ButtonRole.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Mode.Clipboard)
        cb.setText(exception)


def error(message: str, info: str = "", details: str = "") -> None:
    """Shows a little window for error messages."""
    msg = QtWidgets.QMessageBox()
    msg.setIcon(QtWidgets.QMessageBox.Icon.Critical)
    msg.setWindowTitle("Errors occured")
    msg.setText(message)
    if info:
        msg.setInformativeText(info)
    if details:
        msg.setDetailedText(details)
    msg.exec()


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)

# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook
